ini_handler
	New(f)
		if(f)
			Read(f)

	Read(f)
		if(istext(f) && fexists(f))
			f = file(f)
		else if(!fexists(f))
			// In-case the file doesn't exist.
			return 0

		if(isfile(f))
			file = f

		file_text = file2text(file)
		length = length(file_text)

		parse_file()

	proc
		compress_text(string)
			var
				length = length(string)

				start = 1
				end = length - 1
			for(start, text2ascii(string, start) in list(9, 10, 13, 32, 160), start ++)
			for(end, text2ascii(string, end) in list(9, 10, 13, 32, 160), end --)

			return copytext(string, start, end + 1)

		IsValidID(id)
			var
				// A to Z, a to z, 0 to 9, _
				list/characters = list(65, 66, 67, 68, 69, 70, 71, 72, 73,
				74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88,
				89, 90, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107,
				108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119,
				120, 121, 122, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 95)
				length = length(id)

			for(var/a = 1, a <= length, a ++)
				if((a == 1) && (text2ascii(id, a) in (48 to 57)))
					// The first character can't be a number. If it is, then
					// the ID is invalid.
					return 0
				if(!(text2ascii(id, a) in characters))
					// If the character isn't in the valid characters list,
					// then indicate the ID is invalid.
					return 0

			return 1 // ID is valid.

		hex2dec(hex, l)
			hex = uppertext(hex)
			. = 0
			var/length = length(hex)
			for(var/a = 1, a <= length, a ++)
				var/d = text2ascii(hex, a)
				if(!(d in (48 to 57)) && !(d in (65 to 74)))
					Error("Invalid character in hexadecimal field.", l)

				d -= ((d >= 65) ? 55 : 48)
				. += (d * (16 ** (length - a)))

		dec2dec(dec, l)
			// text2num() already exists and does this (except it allows
			// for decimal points and negative numbers, too), but this
			// can check and verify a valid decimal value more easily than
			// using text2num() and a seperate proc.

			. = 0
			var/length = length(dec)
			for(var/a = 1, a <= length, a ++)
				var/d = text2ascii(dec, a)
				if(!(d in (48 to 57)))
					Error("Invalid character in decimal field.", l)

				. += ((d - 48) * (10 ** (length - a)))

		oct2dec(oct, l)
			. = 0
			var/length = length(oct)
			for(var/a = 1, a <= length, a ++)
				var/d = text2ascii(oct, a)
				if(!(d in (48 to 56)))
					Error("Invalid character in octal field.", l)

				. += ((d - 48) * (8 ** (length - a)))

		bin2dec(oct, l)
			. = 0
			var/length = length(oct)
			for(var/a = 1, a <= length, a ++)
				var/d = text2ascii(oct, a)
				if((d != 48) && (d != 49))
					Error("Invalid character in binary field.", l)

				. += ((d - 48) * (2 ** (length - a)))

		parse_file()
			var/n = 1
			for(var/a = n, a <= length, a ++)

				// This causes the loop to stop if the new value of a is
				// less than what it was during the last interation.
				if(a < n)
					break

				n = a

				switch(text2ascii(file_text, a))
					if(35, 59)	// #, ;
						// When finding the start of a comment,
						// automatically jump to the start of the next line.
						var/end = findtext(file_text, "\n", a + 1)
						if(end <= 0)
							break

						a = end

					if(91)		// [
						// This handles figure out the new section, and if
						// the section ID is valid, and all that fun stuff.

						if((a > 1) && (text2ascii(file_text, a - 1) != 10))
							// If the section is not started at the beginning
							// the line, this file is malformed, and send an
							// error.
							Error("Section started in incorrect location.", a)

						var
							end = findtext(file_text, "]", a)
							newline = findtext(file_text, "\n", a)

						if(!end)
							// If no end is found, then the file is malformed
							// and without a definite end to the section header,
							// so cause an error.
							Error("Section header not completed.", a)

						if((newline > 0) && (newline < end))
							// If a newline is found before the section
							// defintion end, cause an error.
							Error("Section header ended in improper location.", a)

						var/section = compress_text(copytext(file_text, a + 1, end + 1))
						if(!IsValidID(section))
							// If the ID is invalid, cause an error.
							Error("Section name is invalid or malformed.", a)

						if(!AddSection(section))
							// Something went wrong, so cause an error.
							Error("Failed to add section to data.", a)

						a = end

					if(65 to 90, 95, 97 to 123)
						// A-Z, _, a-z

						a --
						if(text2ascii(file_text, a) != 10)
							// An attribute has to start at the beginning of the line.
							Error("Attribute started in incorrect location.")

						var
							end = findtext(file_text, "=", a) || findtext(file_text, ":", a)
							newline = findtext(file_text, "\n", a + 1)

						if(isnull(end))
							// If no end is found, then the attribute was not assigned
							// a value, so there's an error.
							Error("Attribute not assigned a value.", a)

						if((newline > 0) && (newline < end))
							// If the newline is found before the end, then the attribte
							// did not have a proper assignment, so there's an error.
							Error("Attribute not assigned a value.", a)

						var/attribute = compress_text(copytext(file_text, a, end))
						if(!IsValidID(attribute))
							// If the ID is invalid, error.
							Error("Attribute name is invalid.", a)

						if(active)
							active.AddAttribute(attribute)

						a = end

					if(34, 39)		// ', "
						// This indicates the beginning of a string.
						var
							string = parse_string(a)
							end
						if(string)
							end = string[2]
							string = string[1]

						if(active && !active.SetValue(string))
							Error("Failed to assign value to attribute.", n)

						a = end

					if(45, 48 to 57)	// -, 0-9
						var
							t = text2ascii(file_text, a + 1)
							end = findtext(file_text, "\n", a + 1)
						if(t in list(98, 111, 100, 120))
							var/r = t
							t = copytext(file_text, a + 2, end)
							switch(r)
								if(98)	// Binary.
									t = bin2dec(t)
								if(111)	// Octal.
									t = oct2dec(t)
								if(100)	// Decimal.
									t = text2num(t)
								if(120)	// Hexadecimal.
									t = hex2dec(t)

						else if(!(t in list(10, 45, 46, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57)) && (end != (n + 1)))
							Error("Invalid character in numeric field.")

						else
							t = text2num(copytext(file_text, a, end))

						if(active && !active.SetValue(t))
							Error("Failed to assign value to attribute.", a)

						a = end

					if(60)	// <
						// This indicates the beginning of a list.
						var/
							list/data = parse_list(n)
							end
						if(data)
							end = data[2]
							data = data[1]

						if(active)
							active.SetValue(data)

						a = end

					if(9, 10, 13, 32, 160)	// Whitespace.
						continue

					else
						Error("Invalid or unknown character occurred during file parsing.", n)

			if(active)
				active.active = null // Null out the active section's active attribute.
				active = null // Null out the active section.

		AddSection(name)
			if(name in sections)
				active = sections[name] // Set this as the active section.
				return -1 // It didn't fail, but it already exists, so return
						  // -1 as a way of indicating this.

			else
				var/ini_handler/section/s = new(src, name)
				active = s // Set it as the active one, for use later.
				sections[name] = s
				return s // Successfully added.

			return 0 // Something failed.

		parse_string(n)
			. = ""
			n ++
			var/m = n
			for(m, m <= length, m ++)
				var/t = text2ascii(file_text, m)
				if(t == 92) // Backslash.
					m ++
					var/sequence = text2ascii(file_text, m + 1)
					if((text2ascii(file_text, m) == 48) && (sequence in list(111, 100, 120)))
						// If the sequence '\0o', '\0d', or '\0x' occurs,
						// it's a character sequence.

						switch(sequence)
							if(111)	// Octal.
								. += ascii2text(oct2dec(copytext(file_text, m + 2, m + 5), m))
							if(100)	// Decimal.
								. += ascii2text(text2num(copytext(file_text, m + 2, m + 5)))
							if(120)	// Hexadecimal.
								. += ascii2text(hex2dec(copytext(file_text, m + 2, m + 5), m))

						m += 4 // Skip ahead four spaces, past the character sequence.
						continue

					if(text2ascii(file_text, m) == 10)
						for(m, m <= length, m ++)
							// If there is a backslash followed by a
							// newline, then keep going until a non-
							// newline character occurs. This is to
							// allow strings that go on multiple
							// lines.
							if(!(text2ascii(file_text, m) in list(9, 10, 13, 32, 160)))
								break

						m --
						continue

					if(text2ascii(file_text, m) == 110)
						. += "\n"
						continue

					if(text2ascii(file_text, m) == 116)
						. += "\t"
						continue

				if(((t == 34) || (t == 39)) && (text2ascii(file_text, m - 1) != 92))
					// If the character is a single or double quote
					// not preceded by a backslash to escape it, it's
					// the end of the string.

					break

				. += copytext(file_text, m, m + 1)

			if(text2ascii(file_text, n - 1) != text2ascii(file_text, m))
				// If the first and last character are not the same
				// (that is, if they are mismatched quotes), then it
				// is an invalid string.
				Error("Mismatched quotes.", n - 1)

			return list(., m)

		parse_list(n)
			n ++

			var/list
				list[0]
				active
				assoc
			for(n, n <= length + 1, n ++)
				var/t = text2ascii(file_text, n)
				switch(t)
					if(34, 39)	// ", '
						if(active && !assoc)
							// If there is an active element, but no
							// indication there should be an associated
							// value, there's an error.
							Error("Extra elements found in single index in list.", n)

						var
							string = parse_string(n)
							end
						if(string)
							end = string[2]
							string = string[1]

						if(active && assoc)
							list[active] = string
						else
							active = string

						n = end

					if(45, 48 to 57)	// -, 0-9
						if(active && !assoc)
							// If there is an active element, but no
							// indication there should be an associated
							// value, there's an error.
							Error("Extra elements found in single index in list.", n)

						var
							conv = 5
							number = null
						for(n, n <= length, n ++)
							var/c = copytext(file_text, n, n + 1)
							switch(c)
								// If the sequence 0b, 0o, 0d, or 0x are
								// found, then this will verify that the
								// sequence is valid and set up the loop
								// to convert the number from one of the
								// radixes and into a decimal number.
								if("b")
									if(number != "0")
										Error("Invalid characters for numeric value.", n)
									conv = 1
									number = null
								if("o")
									if(number != "0")
										Error("Invalid characters for numeric value.", n)
									conv = 1
									number = null
								if("d")
									if(number != "0")
										Error("Invalid characters for numeric value.", n)
									conv = 3
									number = null
								if("x")
									if(number != "0")
										Error("Invalid characters for numeric value.", n)
									conv = 4
									number = null

								if("0", "1")
									// Valid binary, octal, decimal, and
									// hexadecimal characters.
									number += c
								if("2", "3", "4", "5", "6", "7")
									// Valid octal, decimal, and hexadecimal
									// characters.
									if(conv < 1)
										Error("Invalid characters for numeric value.", n)

									number += c
								if("8", "9")
									// Valid decimal and hexadecimal characters.
									if(conv < 2)
										Error("Invalid characters for numeric value.", n)

									number += c

								if("A", "B", "C", "D", "E", "F")
									// Valid hexadecimal characters.
									if(conv != 4)
										Error("Invalid characters for numeric value.", n)

									number += c

								if("-", ".")
									// Valid only for decimal floats. That is, ones
									// not assigned by '0d'.
									if(conv != 5)
										Error("Invalid characters for numeric value.", n)

									number += c

								if(",", ">")
									n --
									break

								else
									if(text2ascii(c) in list(9, 10, 13, 32, 160))
										break

									Error("Invalid characters for numeric value.", n)

						// Conver the string version of the number into the numeric
						// version of it.
						switch(conv)
							if(1) // Binary.
								number = bin2dec(number)
							if(2) // Octal.
								number = oct2dec(number)
							if(3) // Decimal (0d form).
								number = dec2dec(number)
							if(4) // Hexadecimal.
								number = hex2dec(number)
							if(5) // Decimal (float form).
								number = text2num(number)

						if(active && assoc)
							list[active] = number
						else
							active = number

					if(44)	// ,
						if(active && !assoc)
							if(istype(active, /list))
								list[list.len] = active
							else
								list += active

						active = null
						assoc = null

					if(60)	// <
						if(active && !assoc)
							// If there is an active element, but no
							// indication there should be an associated
							// value, there's an error.
							Error("Extra elements found in single index in list.", n)

						// When a list is found, it can not be the index.
						// Therefore, if there is no currently-active
						// element, just add it to the list. If there is
						// one, though, set it as the association of that
						// list.
						var
							list/l = parse_list(n)
							end
						if(l)
							end = l[2]
							l = l[1]
						if(active && assoc)
							list[active] = l
						else
							active = l

						n = end

					if(62)	// >
						// This marks the end of the list, so stop the
						// loop and return what's been found.
						if(active && !assoc)
							if(istype(active, /list))
								list[list.len] = active
							else
								list += active

						active = null
						assoc = null

						break

					if(61, 58)	// =, :
						assoc = 1
						continue

					if(35, 59)	// #, ;
						// When finding the start of a comment,
						// automatically jump to the start of the next line.
						var/end = findtext(file_text, "\n", n + 1)
						if(end <= 0)
							break

						n = end

					if(9, 10, 13, 32, 160)	// Whitespace.
						// There can be an arbitrary amount of whitespace
						// in the list.
						continue

					else
						// Only the above caracters are allowed outside
						// a string, so if something else is found, then
						// there is an error.
						Error("Invalid character in list definition.", n)

			if(text2ascii(file_text, n) != 62)
				// If the list doesn't end with >, it's malformed and
				// therefore in error.
				Error("List malformed.", n)

			return list(list, n)